"""Splunk Integrations."""

from collections import OrderedDict
import datetime
from typing import Any, Union

def splunk_format(**kwargs: Any) -> str:
    """
    Reformat a list of key:value pairs into a simple logging message for Splunk.

    :return: _description_
    :rtype: str
    """
    ordered: OrderedDict[str, Any] = OrderedDict(sorted(kwargs.items()))
    string: list[str] = [f"{str(key)}=\"{value}\"" for key, value in ordered.items()]
    return ','.join(string)


# TODO: fix the abilit to send HEC formats in chuncks.
def splunk_hec_format(host: str, source: str, sourcetype: str,
                      metrics_list: Union[list[str],None] = None,
                      **kwargs: Any) -> dict[str, Any]:
    """Create a JSON style hec format.

    :param host: hostname
    :type host: str
    :param source_name: source of dataa
    :type source_name: str
    :param metrics_list: list of metrics type fileds found in arguments
    :type metrics_list: list[str]
    :param kwargs: key:value pairs to extract and format data structure
    :return: Splunk Hec Datastructure
    :rtype: dict[str,Any]
    """
    hec_json: dict[str, Any] = {
        "time": datetime.datetime.now().timestamp(),
        "host": host,
        "source": source,
        "sourcetype": sourcetype,
        "events": {}
    }
    if metrics_list:
        # Build HEC style Metrics
        hec_json["fields"] = {f"metric_name:{metric}": kwargs.pop(
            metric, None) for metric in metrics_list}
        hec_json["fields"] = dict(sorted(hec_json["fields"].items()))
    hec_json["events"] = {**hec_json["events"], **kwargs}
    hec_json["events"] = dict(sorted(hec_json["events"].items()))
    return hec_json



# pylint: disable=too-many-arguments
"""Decorators."""

from functools import partial
import functools
from inspect import isfunction
import time
import random

from pytoolkit import decorator
from pytoolkit.utils import reformat_exception

def __retry_interval(
        func, exceptions=Exception, tries=-1, delay=0, max_delay=None, backoff=1, jitter=0,
        logger=None):
    """
    Executes a function and retries it if it failed.

    :param func: the funciton to execute.
    :type func: Function
    :param exceptions: an exception or tupple of exceptions to catch, defaults to Exception
    :type exceptions: Exception|tuple[Excpetion,Exception], optional
    :param tries: the maximum number of attempts, defaults to -1 (infinite).
    :type tries: int, optional
    :param delay: intial delay between attempts, defaults to 0.
    :type delay: int, optional
    :param max_delay: the maximum value of delay, defaults to None (no limit).
    :type max_delay: int, optional
    :param backoff: multiplier applied to delay between attempts, defaults to 1 (no backoff).
    :type backoff: int, optional
    :param jitter: extra seconds added to delay between attempts, defaults to 0
                   fixed if a number, random if a tuple (min,max)
    :type jitter: int|tuple[int,int], optional
    :param logger: logger.warning(fmt,error,delay) will be called on failed attempts, defaults to None
                    default is disabled.
    :type logger: Logger, optional
    :return: the result of the func Function.
    """
    _tries, _delay = tries, delay
    while _tries:
        try:
            return func()
        except exceptions as err:
            _tries -= 1
            error = reformat_exception(err)
            if not _tries:
                raise
            if logger is not None:
                logger.warning("msg=\"attempt failed\",error=%s,retrying_in=%ss", error, _delay)
            time.sleep(_delay)
            _delay *= backoff
            if isinstance(jitter, tuple):
                _delay += random.uniform(*jitter)
            else:
                _delay += jitter
            if max_delay is not None:
                _delay = min(_delay, max_delay)


def retry(exceptions=Exception, tries=-1, delay=0, max_delay=None, backoff=1, jitter=0, logger=None):
    """
    Returns a retry decorator.

    :param exceptions: an exception or tupple of exceptions to catch, defaults to Exception
    :type exceptions: Exception|tuple[Excpetion,Exception], optional
    :param tries: the maximum number of attempts, defaults to -1 (infinite).
    :type tries: int, optional
    :param delay: intial delay between attempts, defaults to 0.
    :type delay: int, optional
    :param max_delay: the maximum value of delay, defaults to None (no limit).
    :type max_delay: int, optional
    :param backoff: multiplier applied to delay between attempts, defaults to 1 (no backoff).
    :type backoff: int, optional
    :param jitter: extra seconds added to delay between attempts, defaults to 0
                   fixed if a number, random if a tuple (min,max)
    :type jitter: int|tuple[int,int], optional
    :param logger: logger.warning(fmt,error,delay) will be called on failed attempts, defaults to None
                    default is disabled.
    :type logger: Logger, optional
    :return: a retry decorator.
    :rtype: function
    """
    @decorator
    def retry_decorator(func, *fargs, **fkwargs):
        args = fargs if fargs else list()
        kwargs = fkwargs if fkwargs else dict()
        return __retry_interval(
            partial(func, *args, **kwargs),
            exceptions, tries, delay, max_delay, backoff, jitter, logger)
    return retry_decorator


def __exception_handler(func, exceptions=Exception, default_return=None, message="", logger=None, func_params={}):  # pylint: disable=dangerous-default-value
    """Exception Handler Decorator."""
    try:
        return func()
    except exceptions as err:
        error = reformat_exception(err)
        if logger:
            # need to call func.func to get the original callable function name since created by partial()
            logger.fatal(
                f"function={func.func.__name__},error=\"{message}:error_raw={error}\",level=error")
    if isinstance(default_return, functools.partial):
        return default_return(error=error, level="fatal")
    if default_return:
        return default_return


def error_handler(exceptions=Exception, default_return=None, logger=None, func_params={}):  # pylint: disable=dangerous-default-value
    """
    Error Handler excption; allows passing a default return value if needed.

    :param exceptions: _description_, defaults to Exception
    :type exceptions: _type_, optional
    :param default_return: _description_, defaults to None
    :type default_return: _type_, optional
    :param logger: _description_, defaults to None
    :type logger: _type_, optional
    :param func_params: _description_, defaults to {}
    :type func_params: dict, optional
    :return: _description_
    :rtype: _type_
    """
    @decorator
    def error_handle_decorator(func, *fargs, **fkwargs):
        args = fargs if fargs else list()
        kwargs = fkwargs if fkwargs else dict()
        if isfunction(default_return):
            func_params.update({"func_name": func.__name__})
            func_params.update(kwargs)
            func_params.update({("args" + str(idx+1)): arg for idx, arg in enumerate(args)})
            return __exception_handler(
                partial(func, *args, **kwargs),
                exceptions, partial(default_return, **func_params),
                logger)
        return __exception_handler(
            partial(func, *args, **kwargs),
            exceptions, default_return, logger)
    return error_handle_decorator



"""Files."""

import re
from pathlib import Path
import platform
import tempfile
from typing import Any, Union

import yaml

from pytoolkit.static import ENCODING

# See https://www.linuxtrainingacademy.com/all-umasks/
FILE_UMASK_PERMISSIONS = {
    "default": 0o40775,  # umask 002 rw-rw-r- rwxrwx-r
    "restrictive": 0o40770,  # umask 037 rw-r--- rwxr---
    "root": 0o40755  # umask 022 rw-r-r- rwxr-xr-x
}


def read_yaml(filename: Path) -> dict[str, Any]:
    """
    Read in a YAML configuration file.

    :param filename: Yaml File Full Path
    :type filename: Path
    :return: Yaml Configurations
    :rtype: dict[str,Any]
    """
    with open(filename, 'r', encoding=ENCODING) as r_yaml:
        settings = yaml.safe_load(r_yaml)
    return settings


def get_var_dir(extend_path: Union[str, None] = None, mode: str = "default") -> str:
    """
    Get default var directory depending on OS. Extend to application path if path supplied.
     Allows for user to create a directory strcture where reports or logs can be stored. Using
     a variable directory or extend a path in a users home directory.

    :param extend_path: Extends the system default location;
     creates a new app directory, defaults to None
    :type extend_path: str, optional
    :return: Log Directory for Logs
    :rtype: str
    """
    directory: dict[str, Path] = {
        "darwin": Path.joinpath(Path.home() / "Library/Logs"),
        "linux": Path("/var/log")
    }
    plat: str = platform.system()
    try:
        path = Path(f"{str(directory[plat.lower()])}/{extend_path}") if extend_path else Path(
            str(directory[plat.lower()]))
        mkdir(path=path, mode=mode)
        return str(path)
    except KeyError:
        path = Path(
            f"{str(tempfile.gettempdir())}/{extend_path}") if extend_path else Path(
                tempfile.gettempdir())
        mkdir(path=path, mode=mode)
        return str(path)


def set_location(location: str, extend_path: Union[str, None] = None, mode: str = "default") -> str:
    """Set default logDir or configuration directory based on mode defined."""
    loc = "var"
    if bool([re.match(r'(home|homedir)', r) for r in Path(location).parts]):
        loc = "home"
    mode = mode if mode in FILE_UMASK_PERMISSIONS else "default"
    base_dir = {
        "home": set_homedir,
        "var": get_var_dir
    }
    return base_dir[loc](extend_path=extend_path, mode=mode)


def set_homedir(extend_path: Union[str, None] = None, mode: str = "default") -> str:
    """
    Return the users home dir and can extend if using a subdirctory. Use `mode` to restrict permissions.

    :param extend_path: _description_, defaults to None
    :type extend_path: Union[str,None], optional
    :param mode: _description_, defaults to "default"
    :type mode: str, optional
    :return: _description_
    :rtype: str
    """
    path = Path(Path.home() / extend_path) if extend_path else Path.home()
    mkdir(path=path, mode=mode)
    return str(path)


def mkdir(path: Path, mode: str = "default"):
    """Makes Dir based on permissions passed."""
    #for parent in reversed(path.parents):
    path.mkdir(mode=FILE_UMASK_PERMISSIONS[mode], parents=True, exist_ok=True)


def get_home() -> str:
    """Return current users `home` direcotry."""
    return str(Path.home())


def with_suffix(logName: str) -> str:
    """Add suffix to logname."""
    return str(Path(logName).with_suffix('.log'))



"""Global Static Vars."""

from typing import cast

ENCODING: str = 'utf-8'

NONETYPE: None = cast(None, object())
DEFAULT_TO: list[str] = ['john.doe@acme.com']
DEFAULT_FROM: str = 'python-script@acme.com'
DEFAULT_CC: list[str] = ['']
DEFAULT_BCC: list[str] = ['']
TMP_PEM_POSTFIX = "_pytookit.pem"
PEM_REGEX = r".*[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}_pytoolkit\.pem$"



"""Data Module."""
from dataclasses import dataclass, field

sample_headers = {'test': 1}
@dataclass
class Response:
    body: str
    status: int = field(default=200)
    headers: dict = field(default_factory=dict)
    def __init__(self, body:str, status: int=200):
        self.body = body
        self.status = status
        self.headers = sample_headers



"""mailer"""

from typing import List, Union
import smtplib
import re

from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText

from pytoolkit.utils import string_or_list
from pytoolkit.static import (DEFAULT_TO,DEFAULT_FROM,DEFAULT_CC,DEFAULT_BCC,ENCODING)


def send_mail(smtp_server: str, msg: str = "EMPTY", subject: str = "Python Script",
              mail_to: Union[list[str],str] = DEFAULT_TO,
              mail_from: str = DEFAULT_FROM,
              mail_cc: Union[list[str],str] = DEFAULT_CC,
              mail_bcc: Union[list[str],str] = DEFAULT_BCC,
              msg_html: Union[str,None] = None,
              attachment: Union[str,None] = None, port: int = 25):
    """Send Mail

    :param smtp_server: _description_
    :type smtp_server: str
    :param msg: _description_, defaults to "EMPTY"
    :type msg: str, optional
    :param subject: _description_, defaults to "Python Script"
    :type subject: str, optional
    :param mail_to: _description_, defaults to _default_to
    :type mail_to: list, optional
    :param mail_from: _description_, defaults to _default_from
    :type mail_from: str, optional
    :param mail_cc: _description_, defaults to _default_cc
    :type mail_cc: list, optional
    :param mail_bcc: _description_, defaults to _default_bcc
    :type mail_bcc: list, optional
    :param msg_html: _description_, defaults to None
    :type msg_html: str, optional
    :param attachment: _description_, defaults to None
    :type attachment: str, optional
    :param port: _description_, defaults to 25
    :type port: int, optional
    :return: _description_
    :rtype: _type_
    """
    # Handle if a string is passed
    delimeters = ',| '
    mail_to: List[str] = string_or_list(value=mail_to,delimeters=delimeters)
    mail_cc: List[str] = string_or_list(value=mail_cc,delimeters=delimeters)
    mail_bcc: List[str] = string_or_list(value=mail_bcc,delimeters=delimeters)

    message = MIMEMultipart('alternative')
    message['Subject'] = subject
    message['From'] = mail_from
    message['To'] = ", ".join(mail_to)
    message['cc'] = ", ".join(mail_cc)
    message['bcc'] = ", ".join(mail_bcc)

    try:
        if attachment and isinstance(attachment, list):
            for attach in attachment:
                # add file as application/octet-stream
                # email client can usually downlaoad this automatically as an attachemment
                part = MIMEBase('application', 'octet-stream')
                with open(attach, 'rb', encoding=ENCODING) as a:
                    # Add encoded stream to MIMEBase app payload
                    part.set_payload(a.read())
                    # encode
                    encoders.encode_base64(part)
                # Get Filename
                split: list[str] = re.findall(r"[\w']+", attach)
                filename: str = f'{split[-2]}.{split[-1]}'
                # Add header as key/value pair to attach part
                part.add_header(
                    "Content-Disposition",
                    f"attachment; filename={filename}",
                )
                # Add attachment to message and convert message to string
                message.attach(part)
        if msg_html:
            html_email_message = MIMEText(msg_html, 'html')
            message.attach(html_email_message)
        else:
            plaintext_email_message = MIMEText(msg, 'plain')
            message.attach(plaintext_email_message)
        # Send Email
        with smtplib.SMTP(smtp_server, port) as server:
            server.ehlo()
            server.send_message(message)
            server.quit()
        return "Success"
    except (smtplib.SMTPConnectError, smtplib.SMTPDataError, smtplib.SMTPAuthenticationError, smtplib.SMTPHeloError, smtplib.SMTPServerDisconnected) as err:
        return f"SMTP Connection Error: {str(err)}"
    except (smtplib.SMTPException, smtplib.SMTPSenderRefused, smtplib.SMTPNotSupportedError, smtplib.SMTPResponseException, smtplib.SMTPRecipientsRefused) as err:
        return f"SMTP Communication Error: {str(err)}"
    except Exception:
        return "SMTP Unknown Error"



# pylint: disable=line-too-long
"""Package Supplied Utilities."""

from typing import Any, Generator, Hashable, Union
from collections.abc import MutableMapping

from dataclasses import dataclass, fields
import pandas as pd

from pytoolkit.static import NONETYPE


@dataclass
class BaseMonitor:
    """Base Dataclass Methods."""
    @classmethod
    def create_from_dict(cls, dict_: dict[str, Any]):
        """
        Class Method that returns dataclass using a
         dictionary and strips invalid params.

        :param dict_: _description_
        :type dict_: _type_
        :return: Dataclass
        :rtype: :dataclass: DataModel
        """
        class_fields = {f.name for f in fields(cls)}
        return cls(**{k: v for k, v in dict_.items() if k in class_fields})

    @classmethod
    def create_from_kwargs(cls, **kwargs: Any):
        """
        Class method that returns dataclass values by
          unpacking paramter values and strips out invalid params.

        :param kwargs: unpacked key:value pairs
        :type **kwargs: variable length argument list
        :return: Dataclass
        :rtype: :dataclass: DataModel
        """
        class_fields = {f.name for f in fields(cls)}
        return cls(**{k: v for k, v in kwargs.items() if k in class_fields})

    def to_dict(self, extend: bool = True):
        """
        Returns dataclass as dictionary.

        :param extend: Extends the dataclass that have a value `NONETYPE`.
        :type bool: Optional True
        :return: dataclass dictionary
        :rtype: dict[str, Any]
        """
        return {key: value for key, value in self.__dict__.items() if value is not NONETYPE}


def _flatten_dict_gen(_d: MutableMapping[str, Any], parent_key: str,
                      sep: str, extended_label: bool,
                      skip_item: list[str]) -> Generator[tuple[str, Any], Any, None]:
    for k, v in _d.items():
        new_key: str = k
        if extended_label:
            new_key: str = parent_key + sep + \
                k if (parent_key and k not in skip_item) else k  # type: ignore
        if isinstance(v, MutableMapping):
            yield from flatten_dict(v,  # type: ignore
                                    new_key, extended_label=extended_label,
                                    skip_item=skip_item,
                                    sep=sep).items()
        else:
            yield new_key, v


def flatten_dict(_dict: MutableMapping[str, Any], parent_key: str = "",
                 sep: str = ".", extended_label: bool = True,
                 skip_item: Union[list[str], None] = None) -> dict[str, Any]:
    """
    Flatten out a dictionary with nested values.

    :param _dict: Dictionary
    :type _dict: MutableMapping[str,Any]
    :param parent_key: Top Level Key, defaults to ""
    :type parent_key: str, optional
    :param sep: Seperator, defaults to "."
    :type sep: str, optional
    :param extended_label: Uses the same key by default or appends the hierarchy
     into the name of the key used to express the nesting structure, defaults to True
    :type extended_label: bool, optional
    :param skip_item: List of Keys to ignore and flatten without the parent, defaults to []
    :type skip_item: list, optional
    :return: Flattened Dictionary
    :rtype: dict[str,Any]
    """
    skip_item = skip_item if skip_item else [""]
    return dict(_flatten_dict_gen(_dict, parent_key, sep, extended_label, skip_item))


def flatten_dictionary(_dict: MutableMapping[Any, Any], sep: str = '.') -> dict[Hashable, Any]:
    """
    Flatten a dictionary via pandas normalizer.

    :param d: _description_
    :type d: MutableMapping
    :param sep: Adds seperator to Key, defaults to '.'
    :type sep: str, optional
    :return: Flattened Dictionary.
    :rtype: MutableMapping
    """
    [flat_dict] = pd.json_normalize(_dict, sep=sep).to_dict(orient='records')  # type: ignore
    return flat_dict

# TODO: fix the nested structure add abiltiy to read in a csv or XCEL to help maniplate proper csv human readable datastructures


def _nest_dict_rec(key: str, value: Any, sep: str, out: dict[str, Any]) -> None:
    """Maniplates dictionary into a nested dictionary based on seperator."""
    key, *rest = key.split(sep, 1)
    if rest:
        _nest_dict_rec(rest[0], value, sep, out.setdefault(key, {}))
    else:
        out[key] = value


def nested_dict(_dict: MutableMapping[str, Any], sep: str = '.') -> dict[str, Any]:
    """
    Transform a Flattened Dictionary into a Nested Dictionary.

    :param _dict: _description_
    :type _dict: MutableMapping[str, Any]
    :param sep: Seperator to split Keys with, defaults to '.'
    :type sep: str, optional
    :return: Nested Dictionary.
    :rtype: dict[str, Any]
    """
    result: dict[str, Any] = {}
    for k, v in _dict.items():
        _nest_dict_rec(k, v, sep, result)
    return result
